/*------------------------------------------------------------------------------
  Reflash tool for Carberry Project
  www.carberry.it
  Massimo Savina
------------------------------------------------------------------------------*/
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h> 
#include <time.h>

#include "reflash.h"

#define CARBERRY_ADDR "127.0.0.1"
#define CARBERRY_PORT 7070
#define ERASEUP "\033[2K\033[1A"

#define true 1
#define false 0

int sockfd = 0;

/*------------------------------------------------------------------------------
------------------------------------------------------------------------------*/
int CarberryConnect()
{
  struct sockaddr_in serv_addr;

  while (true)
  {
    printf("Connecting Carberry Server...\r\n");

    if (sockfd)
    {
      close(sockfd);
      sockfd = 0;
    }

    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0)
    {
      printf("Error: Could not create socket\r\n");
      sleep(1);
      continue;
    }

    memset(&serv_addr, '0', sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(CARBERRY_PORT);

    if (inet_pton(AF_INET, CARBERRY_ADDR, &serv_addr.sin_addr) <= 0)
    {
      printf("Error: inet_pton error occurred\r\n");
      sleep(1);
      continue;
    }

    if (connect(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0)
    {
       printf("Carberry Server Connection Failed. Sleep and retry\r\n");
       sleep(1);
       continue;
    }
    else
    {
      //Ok!
      printf("Carberry Ready\r\n");
      return true;
    }
  }
}

/*------------------------------------------------------------------------------
  Send a single line (CR/LF terminated) to socket
------------------------------------------------------------------------------*/
int SendSock(int* fd, char* buffer, unsigned int len)
{
  //Send to socket
  int sent = send(*fd, buffer, len, 0);
  if (sent <= 0)
  {
    //Maybe server has closed connection
    *fd = 0;
    return false;
  }
  else
  {
    return true;
  }
}

/*------------------------------------------------------------------------------
  Receive a single line (CR/LF terminated) from socket
  Timeout of to sec
------------------------------------------------------------------------------*/
int RecvSock(int* fd, char* buffer, unsigned int maxlen, unsigned int to)
{
  int i = 0;
  int j = 0;
  int nbytes = 0;
  struct timeval timeout;
  char tmp[STDSTR];
  fd_set reading;

  while (true)
  {
    FD_ZERO(&reading);
    FD_SET(*fd, &reading);
    //timeout
    timeout.tv_sec  = to;
    timeout.tv_usec = 0;

    //Waiting for something
    switch (select(*fd+1, &reading, NULL, NULL, &timeout))
    {
      //Error
      case -1:
        buffer[i] = 0;
        return false;
      break;
      //Timeout
      case 0:
        buffer[i] = 0;
        return false;
      break;
      //Data
      default:
        if (FD_ISSET(*fd, &reading))
        {
          //Get socket chars
          nbytes = recv(*fd, tmp, STDSTR, 0);
          //Error or closed
          if (nbytes <= 0)
          {
            *fd = 0;
            return false;
          }

          for (j=0; j<nbytes; j++)
          {
            //Overflow
            if (i == (maxlen-1)) return false;
            //Collect new char
            buffer[i++] = tmp[j];
            if (tmp[j] == '\n')
            {
              buffer[i++] = 0;
              return true;
            }
          }
        }
      break;
    }
  }
}

/*------------------------------------------------------------------------------
------------------------------------------------------------------------------*/
int Talk(const char *format, ...)
{
  unsigned int lines = 1000;
  unsigned int tries = 3;
  char line[STDSTR];
  va_list ap;

  while (tries--)
  {
    //Prepare command
    va_start(ap, format);
    vsprintf(line, format, ap);
    va_end(ap);

    //Send command to server
    if (!SendSock(&sockfd, line, strlen(line)))
    {
      //Debug
      printf("Send Error!\r\n");
      CarberryConnect();
      continue;
    }
    while (lines--)
    {
      //Waiting for reply
      if (!RecvSock(&sockfd, line, STDSTR, 5))
      {
        //Debug
        printf("Recv Error!\r\n");
        if (!sockfd) CarberryConnect();
        break;
      }
      //Parse reply
      if      (strstr(line, "OK"))
      {
        return true;
      }
      else if (strstr(line, "ERROR"))
      {
        return false;
      }
      else
      {
        //Get another line
      }
    }
  }
  return false;
}

/*------------------------------------------------------------------------------
------------------------------------------------------------------------------*/
void Reflash(char* fname)
{
  FILE* fp;
  char cmd[128];
  char tmp[64];
  char buffer[64];
  int i;
  long lSize;
  long writed = 0;
  size_t bytes;

  if (!sockfd)
  {
    return;
  }

  //Open upgrade file
  printf("Opening upgrade file...\r\n");
  fp = fopen(fname, "rb");
  if (!fp)
  {
    //File not exists
    printf("Unable to open %s\r\n", fname);
    return;
  }

  //Get file size
  fseek(fp, 0, SEEK_END);
  lSize = ftell(fp);
  rewind(fp);

  Talk("AT\r\n");

  printf("Probing boot mode...");
  fflush(stdout);
  if (!Talk("BOOT\r\n"))
  {
	  printf("\r\n");
	  printf("Entering Carberry in boot mode...");
	  fflush(stdout);
	  if (!Talk("REFLASH\r\n"))
	  {
		printf("\r\nError in REFLASH command\r\n");
		return;
	  }
	  printf("Ok\r\n");

	  //Wait for a bit
	  sleep(5);

	  printf("Probing boot mode...");
	  fflush(stdout);
	  if (!Talk("BOOT\r\n"))
	  {
		printf("\r\nBOOT mode not ready\r\n");
		return;
	  }
	  printf("Ok\r\n");
  }
  else
  {
	printf("Ok\r\n");
  }


  printf("Erasing flash...");
  fflush(stdout);
  if (!Talk("ERASE\r\n"))
  {
	printf("\r\ERROR in ERASE command\r\n");
	return;
  }
  printf("Ok\r\n");


  //Write
  do	
  {
    //Read some data
    bytes = fread(buffer, 1, 52, fp);
    if (bytes)
    {
      sprintf(cmd, "WRITE ");
      
      for (i=0; i<52; i++)
      {
        sprintf(tmp, "%c",  buffer[i]);
        strcat(cmd, tmp);
      }
      sprintf(tmp, "\r\n");
      strcat(cmd, tmp);

      //Debug! printf(cmd);
      if (!Talk(cmd))
      {
   	    printf("Write error!\r\n");
	    return;
	  }
	  else
	  {
	    writed += 52;
	    printf("Writing... %0.1f%%\r\n%s", ((float)writed/(float)lSize*(float)100), ERASEUP);
	  }
    }
  }
  while (bytes);
  fclose(fp);
  printf("\r\n");
  fflush(stdout);
        
  printf("Starting Application...");
  fflush(stdout);
  if (!Talk("START\r\n"))
  {
	printf("\r\ERROR in START command\r\n");
	return;
  }
  printf("Ok\r\n");
  printf("Done!\r\n");
  return;
}


/*------------------------------------------------------------------------------
------------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
    //Connecting to Carberry Server
    CarberryConnect();

    if (argc < 1)
    {
      printf("Usage: reflash <file_to_reflash>");
      return 0;
    }
    
    Reflash(argv[1]);

    return 0;
}
